import { __awaiter } from 'tslib';
import { auth, firestore, storage } from 'firebase';
import { Injectable, NgZone, NgModule } from '@angular/core';
import { HttpClient, HttpClientModule } from '@angular/common/http';
import { CommonModule } from '@angular/common';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 *
 * @file library.ts
 *
 */
class Library {
    /**
     * Returns http query string.
     * @param {?} params Object to build as http query string
     * @return {?}
     *      - http query string
     *      - Or null if the input is emtpy or not object.
     */
    static httpBuildQuery(params) {
        if (Library.isEmpty(params)) {
            return null; //
        }
        const /** @type {?} */ keys = Object.keys(params);
        if (keys.length === 0) {
            return null; //
        }
        const /** @type {?} */ esc = encodeURIComponent;
        const /** @type {?} */ query = keys
            .map(k => esc(k) + '=' + esc(params[k]))
            .join('&');
        return query;
    }
    /**
     * Returns n'th portion of the input `str` after spliting by the `separator`
     *
     * \@code
     *      const str = 'abc.def.ghi';
     *      return this.library.segment( str, '.', 0 ); // returns `abc`
     *
     * @param {?} str string to get a portion from.
     * @param {?=} separator to split the string. Default is a Blank.
     * @param {?=} n n'th portion to get. Index begins with 0. Default is 0.
     * @return {?}
     *      - a portion of the input string.
     *      - or null
     *          - if the input `str` is empty.
     *          - if the input `str` is not a string.
     *          - if the n'th portion does not exists.
     *          - if the value of the portion is empty
     *          - if separator is not a string and empty.
     *
     */
    static segment(str, separator = ' ', n = 0) {
        if (typeof str !== 'string') {
            return null;
        }
        if (typeof separator !== 'string' || !separator) {
            return null;
        }
        if (str) {
            const /** @type {?} */ re = str.split(separator);
            if (re[n] !== void 0 && re[n]) {
                return re[n];
            }
        }
        return null;
    }
    /**
     * Returns true if the input `what` is falsy or empty or no data.
     * @param {?} what
     * @return {?} true if the input `what` is
     *          - falsy value.
     *              -- boolean and it's false,
     *              -- number with 0.
     *              -- string with empty. ( if it has any vlaue like blank, then it's not empty. )
     *              -- undefined.
     *          - object with no key.
     *          - array with 0 length.
     *
     *      - otherwise return false.
     */
    static isEmpty(what) {
        if (!what) {
            return true; // for number, string, boolean, any falsy.
        }
        if (typeof what === 'object') {
            return Object.keys(what).length === 0;
        }
        if (Array.isArray(what)) {
            return what.length === 0;
        }
        return false;
    }
    /**
     * Returns true if the input `a` and `b` are identical.
     * @param {?} a Object a
     * @param {?} b Object b
     * @return {?}
     */
    static isEqual(a, b) {
        if (typeof a === 'object' && typeof b === 'object') {
            const /** @type {?} */ aKeys = Object.keys(a);
            const /** @type {?} */ bKeys = Object.keys(b);
            if (aKeys.length !== bKeys.length) {
                return false;
            }
            return aKeys.findIndex((v, i) => v !== bKeys[i]) === -1;
        }
        else if (Array.isArray(a) && Array.isArray(b)) {
            if (a.length !== b.length) {
                return false;
            }
            else {
            }
        }
        else {
            return a === b;
        }
    }
    /**
     * @param {?} str
     * @return {?}
     */
    static isString(str) {
        return typeof str === 'string';
    }
    /**
     *
     * Removes properties with `undefined` value from the object and returns it.
     *
     * You cannot set `undefiend` value into firestore `document`. It will produce a Critical error.
     *
     * @param {?} obj Object to be set into `firestore`.
     *      It is passed by reference.
     *
     * @return {?} the input object that has sanitized.
     */
    static sanitize(obj) {
        if (obj) {
            if (typeof obj === 'object') {
                Object.keys(obj).forEach(key => obj[key] === undefined && delete obj[key]);
            }
        }
        /** Remove `password` not to save on documents. */
        if (obj && obj['password'] !== void 0) {
            delete obj['password'];
        }
        return obj;
    }
    /**
     * Removes space(s) between the separator in `separator`
     * \@description
     *      If the input str is given with `a, b, c c ,d `, then the return will be `a,b,c c,d`.
     * @param {?} separator separator in string
     * @param {?} str string to remove space from.
     *
     * @return {?} a string after removing spaces between the `separator`.
     *      - if the string is falsy, it returns the input `str` itself.
     */
    static removeSpaceBetween(separator, str) {
        if (!str) {
            return str;
        }
        else {
            return str.split(separator).map(s => s.trim()).join(separator);
        }
    }
    /**
     * Returns browser language
     *
     * @param {?=} full If it is true, then it returns the full language string like 'en-US'.
     *              Otherwise, it returns the first two letters like 'en'.
     *
     * @return {?}
     *      - the browser language like 'en', 'en-US', 'ko', 'ko-KR'
     *      - null if it cannot detect a language.
     */
    static getBrowserLanguage(full = false) {
        const /** @type {?} */ nav = window.navigator;
        const /** @type {?} */ browserLanguagePropertyKeys = ['language', 'browserLanguage', 'systemLanguage', 'userLanguage'];
        let /** @type {?} */ ln = null;
        // support for HTML 5.1 "navigator.languages"
        if (Array.isArray(nav.languages)) {
            for (let /** @type {?} */ i = 0; i < nav.languages.length; i++) {
                const /** @type {?} */ language = nav.languages[i];
                if (language && language.length) {
                    ln = language;
                    break;
                }
            }
        }
        // support for other well known properties in browsers
        for (let /** @type {?} */ i = 0; i < browserLanguagePropertyKeys.length; i++) {
            const /** @type {?} */ language = nav[browserLanguagePropertyKeys[i]];
            if (language && language.length) {
                ln = language;
                break;
            }
        }
        if (ln) {
            if (full === false) {
                ln = ln.substring(0, 2);
            }
        }
        return ln;
    }
    /**
     *
     * Returns a string after patching error information.
     * \@code
     *      _.patchmarker( 'Unknown #no', {no: 123} ) // returns 'Unknown 123'
     *
     * @param {?} str Error string
     * @param {?=} info Error information to patch into the string
     *
     *
     *
     * @return {?} patched string
     *
     */
    static patchMarker(str, info = null) {
        if (info === null || typeof info !== 'object') {
            return str;
        }
        const /** @type {?} */ keys = Object.keys(info);
        if (!keys.length) {
            return str;
        }
        for (const /** @type {?} */ k of keys) {
            str = str.replace('#' + k, (/** @type {?} */ (info[k])));
        }
        return str;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
const COLLECTIONS = {
    USERS: 'users',
    SETTINGS: 'settings',
    CATEGORIES: 'categories',
    POSTS: 'posts',
    // POSTS_DELETED: 'posts_deleted', // no more in use.
    /**
         * Sub collections. No prefix for subcollections.
         */
    LIKES: 'likes',
    DISLIKES: 'dislikes',
    COMMENTS: 'comments'
};
const POST_DELETED = 'This post has been deleted!';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 *
 * @file error.ts
 * @description It only holds constants. No class, functions, vars.
 *          - the value of the constant must be a string only. It cannot contain numeric value.
 */
const UNKNOWN = 'unknown-error';
const FIREBASE_API_ERROR = 'firebase-api-error';
const NOT_FOUND = 'not-found';
const ALREADY_EXISTS = 'already-exists';
const PERMISSION_DENIED = 'permission-denied';
const RESOURCE_EXHAUSTED = 'resource-exhausted';
const USER_IS_NOT_LOGGED_IN = 'You are not logged in';
const NO_DOCUMENT_ID = 'no-document-id';
const DOCUMENT_ID_TOO_LONG = 'document-id-too-long';
const DOCUMENT_ID_CANNOT_CONTAIN_SLASH = 'document-id-cannot-cotain-slash';
const COLLECTION_NOT_EMPTY = 'collection-not-empty';
const CATEGORY_ID_EMPTY = NO_DOCUMENT_ID;
const CATEGORY_EXISTS = ALREADY_EXISTS;
const CATEGORY_DOES_NOT_EXIST = NOT_FOUND;
const POST_ID_NOT_EMPTY = 'post-id-not-empty';
const POST_ID_EMPTY = 'post-id-empty';
const ALREADY_LIKED = 'already-liked-or-disliked';
// USER
// firebase errors
const EMAIL_ALREADY_IN_USE = 'auth/email-already-in-use';
const USER_NOT_FOUND = 'auth/user-not-found';
const INVALID_EMAIL = 'auth/invalid-email';
const WEAK_PASSWORD = 'auth/weak-password';
const EXPIRED_ID_TOKEN = 'auth/id-token-revoked';
const WRONG_PASSWORD = 'auth/wrong-password';
// firelibrary errors
const PASSWORD_TOO_LONG = 'password-too-long';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
const en = {
    'HOME': 'Home'
};
/**
 * Error Code and Text
 */
en[UNKNOWN] = 'Unkown error. #info';
en[FIREBASE_API_ERROR] = 'Firebase API error. #info';
en[CATEGORY_ID_EMPTY] = 'Category ID is empty.';
en[NOT_FOUND] = 'Document was not found. Document ID: #documentID';
en[ALREADY_EXISTS] = 'Category is already exists. Category ID: #categoryID';
en[DOCUMENT_ID_TOO_LONG] = 'Document ID is too long. Document ID: #documentID';
en[PERMISSION_DENIED] = 'Permission denied or insufficient permission. #info';
en[POST_ID_EMPTY] = 'Post ID is empty.';
en[POST_ID_NOT_EMPTY] = 'Post ID is not empty.';
// USER
en[EMAIL_ALREADY_IN_USE] = 'Email address already in use. #info';
en[USER_NOT_FOUND] = 'User was not found in collection. #info';
en[INVALID_EMAIL] = 'Invalid email, #info';
en[WEAK_PASSWORD] = 'Weak password, #info';
en[PASSWORD_TOO_LONG] = 'Password length exceeds at maximum 128 characters.';
en[EXPIRED_ID_TOKEN] = 'ID token is revoked/expired. Login again to get a new id token.';
en[WRONG_PASSWORD] = 'Password was incorrect, #info';
en[USER_IS_NOT_LOGGED_IN] = 'You are not logged in.';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * @file settings.ts
 * @description This script file holds the settings of firelibrary.
 *
 */
let domain;
domain = 'localhost';
/**
 * Root collection name for `FireLibrary` documents and collections.
 * \@warninig if you change `root collection name`, you need to rewrite the security rules.
 */
const COLLECTION_ROOT = 'fire-library';
/**
 * Prefix for collection
 * \@todo If you change the prefix, you need to update `Firestore Rules` accordingly.
 *
 */
const COLLECTION_DOMAIN = domain;

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class Base {
    /**
     * @param {?=} collectionName
     */
    constructor(collectionName = '') {
        this.collectionName = collectionName;
        // static ngZone;
        this.auth = null;
        this.db = null;
        this.storage = null;
        this.auth = Base.firebase.auth();
        this.db = Base.firebase.firestore();
        this.storage = Base.firebase.storage();
    }
    /**
     * @param {?} config
     * @return {?}
     */
    static configure(config) {
        Base.firebase = config.firebaseApp;
        Base.functions = config.functions;
    }
    /**
     * @return {?}
     */
    version() {
        return '0.0.2';
    }
    /**
     * @return {?}
     */
    get settings() {
        return Base.settings;
    }
    /**
     * @return {?}
     */
    get collectionRoot() {
        return Base.collectionRoot;
    }
    /**
     * @return {?}
     */
    get collectionDomain() {
        return Base.collectionDomain;
    }
    /**
     * @param {?} obj
     * @return {?}
     */
    setSettings(obj) {
        if (obj) {
            Base.settings = Object.assign(Base.settings, obj);
        }
    }
    /**
     * @return {?}
     */
    get http() {
        return Base.http;
    }
    /**
     * Returns the collection of the selected collection in this.collectionName
     * @return {?}
     */
    get collection() {
        if (this.collectionName) {
            // console.log('col name: ', this.collectionName);
            return this.db.collection(this.collectionRoot)
                .doc(this.collectionDomain)
                .collection(this.collectionName);
        }
        else {
            return null;
        }
    }
    /**
     * Returns a collection reference under the domain.
     * @desc do not use this.db.collection(....)
     *          Use this method to get a collection reference.
     * @param {?} name
     * @return {?} a collection reference.
     *
     */
    collectionRef(name) {
        if (name) {
            return this.db.collection(this.collectionRoot)
                .doc(this.collectionDomain)
                .collection(name);
        }
        else {
            return null;
        }
    }
    /**
     * @return {?}
     */
    get settingsReference() {
        return this.collectionRef(COLLECTIONS.SETTINGS);
    }
    /**
     *
     * Returns an object of RESPONSE interface.
     *
     * @param {?=} data
     * @return {?}
     *      - <RESPONE> object.
     */
    success(data) {
        return {
            code: null,
            data: data
        };
    }
    /**
     *
     * Returns a Promise.
     *
     * @desc This method returns a Promise. And this is important to understand.
     *      Firstly, it will first return a Promise.
     *      Secondl, laster it will reject().
     *
     * @desc This method is actually an Asynchronous method since it returns a Promise.
     *
     * @desc `_count` counts how many times the `Error Object` passed over to this method.
     *          - when an Error that has already done translate with `failture()`, the does not translate again.
     *              it simply increase `_count`.
     *              so, it is safe to pass one object to this method over gain.
     *
     *
     * \@code this.failure( UNKNOWN );
     * \@code this.failure( new Error(UNKNOWN) ); // same as above.
     *
     * \@example test/test.error.ts
     * @see test.error.ts on how it throws error.
     *
     * @param {?} obj error object or error code string.
     *      If the input `obj` is a string, then it encapsulates with ` new Error( obj ) `.
     *      So, the input `obj` is transformed into an Error object.
     *      If the input `obj` is an Error object, It may be a Firebase Error.
     *      It will detect if it's a Firebase Error and translates the Firebase Error into proper `error message`.
     *          ( The error code of Firebase Error should not be changed. Only error message should be translated. )
     *      Unfortunately we cannot use instance of `firebase.firestore.FirestoreError` or we do not know how to use it yet.
     *      So, we simply compare if the error code is one of the firebase error code.
     *      \@see for Firestore, https://firebase.google.com/docs/reference/js/firebase.firestore.FirestoreError
     *      \@see for Authentication, https://firebase.google.com/docs/reference/js/firebase.auth.Error
     *
     * @param {?=} info
     * @return {?} Promise<never>
     *
     */
    failure(obj, info = {}) {
        let /** @type {?} */ e;
        if (typeof obj === 'string') {
            e = new Error(obj);
        }
        else {
            e = obj;
        }
        if (e['_count']) {
            // Already translated.
            e['_count']++;
            return Promise.reject(e);
        }
        if (this.isFirebaseError(e, info)) {
            this.translateFirebaseError(e, info);
        }
        else {
            e['code'] = e.message;
            e['message'] = this.translate(e.message, info);
        }
        if (e['code'] === e['message']) {
            e['message'] = `Error code - ${e['code']} - is not translated. Please translate it. It may be firebase error.`;
        }
        e['_count'] = 1;
        return Promise.reject(e);
    }
    /**
     * Returns true if the error is a Firebase Error object.
     *
     * \@description
     *      - It checks if the error code is the same as `Firebase Error Code`. If yes, it returns true.
     * @param {?} e
     * @param {?} info
     * @return {?}
     */
    isFirebaseError(e, info) {
        // console.log('error: ', e.code, e.message);
        switch (e.code) {
            case 'not-found':
                // console.log('not-found: ', e.message);
                const /** @type {?} */ str = e.message;
                info['documentID'] = str.split('/').pop();
                return true;
            case INVALID_EMAIL:
            case WRONG_PASSWORD:
            case PERMISSION_DENIED:
            case ALREADY_EXISTS:
            case EMAIL_ALREADY_IN_USE:
            case WEAK_PASSWORD:
            case USER_NOT_FOUND:
            case EXPIRED_ID_TOKEN:
            case PASSWORD_TOO_LONG:
            case USER_IS_NOT_LOGGED_IN:
                info['info'] = e.message;
                return true;
            default:
                return false;
        }
    }
    /**
     * @param {?} e
     * @param {?} info
     * @return {?}
     */
    translateFirebaseError(e, info) {
        e['message'] = this.translate(e.code, info);
    }
    /**
     * @param {?} code
     * @param {?=} info
     * @return {?}
     */
    translate(code, info) {
        return Library.patchMarker(this.getText(code), info);
    }
    /**
     * Returns the text of the code.
     * @desc If the language is not `en`, then it gets the text of the language.
     *
     * @param {?} code
     * @return {?} text of that code.
     *      - if the code does not exist on text file, then it returns the code itself.
     */
    getText(code) {
        const /** @type {?} */ ln = this.getLanguage();
        let /** @type {?} */ text = null;
        // console.log('getText: ', ln, Base.texts);
        if (this.getLanguage() !== 'en') {
            // if not English,
            if (Base.texts[ln] !== void 0 && Base.texts[ln][code] !== void 0) {
                // check if the text of the language exists
                text = Base.texts[ln][code];
            }
        }
        if (!text) {
            // if it's not English or the text of that language not found,
            if (Base.texts['en'][code] !== void 0) {
                // get the text of the code in English
                text = Base.texts['en'][code];
            }
        }
        if (!text) {
            // if no text found, return the code.
            text = code;
        }
        return text;
    }
    /**
     * @return {?}
     */
    getLanguage() {
        return Base.language;
    }
    /**
     * Sets a language and loads the language file.
     * This will load JSON language file under `assets/lang` by default. You can change the path.
     * @desc If the input `ln` is 'en', then it will just return since `en` language is loaded by typescript by default.
     * @desc If the language is already loaded, it does not load again.
     *
     * @see README## Langulage Translation
     *
     * \@code
     *          fire.setLanguage('cn')
     * .catch( e => alert('Failed to load language file. ' + e.message) );
     *
     * @param {?} ln
     * @return {?} a Promise<any> on success. Otherwise error will be thrown.
     */
    setLanguage(ln) {
        Base.language = ln;
        if (ln === 'en') {
            return Promise.resolve();
        }
        if (Base.texts[ln]) {
            return Promise.resolve();
        }
        return this.http.get(`/${Base.languageFolder}/${ln}.json`).toPromise()
            .then(re => Base.texts[ln] = re);
    }
    /**
     *
     * Checks if the Document ID is in right format.
     *
     *
     * @param {?} documentID Document ID
     * @return {?} null if there is no problem. Otherwise `ERROR CODE` will be returned.
     *
     *
     *
     */
    checkDocumentIDFormat(documentID) {
        if (Library.isEmpty(documentID)) {
            return NO_DOCUMENT_ID;
        }
        if (documentID.length > 128) {
            return DOCUMENT_ID_TOO_LONG;
        }
        if (documentID.indexOf('/') !== -1) {
            return DOCUMENT_ID_CANNOT_CONTAIN_SLASH;
        }
        return null;
    }
    /**
     * Get a document.
     * @desc This is a general method to get a document.
     *      - It can be overriden on each module.
     *
     * @param {?} id
     * @return {?} Promise<Document Data> or Error is thrown.
     *      If there is no document by that id, then NOT_FOUND errir will be thrown.
     */
    getValidator(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const /** @type {?} */ idCheck = this.checkDocumentIDFormat(id);
            if (idCheck) {
                return this.failure(new Error(idCheck), { documentID: id });
            }
            return null;
        });
    }
    /**
     * @param {?} id
     * @return {?}
     */
    get(id) {
        return this.getValidator(id)
            .then(() => {
            return this.collection.doc(id).get();
        })
            .then(doc => {
            if (doc.exists) {
                return this.success(doc.data());
            }
            else {
                return this.failure(new Error(NOT_FOUND));
            }
        })
            .catch(e => this.failure(e));
    }
    /**
     * Returns the document of likes/dislikes.
     *
     * @desc This gets `document of like/dislike` and used by post/comment.
     * @param {?} collectionRef
     * @return {?}
     */
    likeDocument(collectionRef) {
        /**
         * \@todo needs to improve here. It shouldn't use `firebase` directly. It needs to refactor the structure.
         */
        const /** @type {?} */ uid = auth().currentUser.uid;
        const /** @type {?} */ ref = collectionRef.doc(uid);
        console.log(`likeDocument path: `, ref.path);
        return ref;
    }
    /**
     * This does validation for `like`, `unlike`, `dislike`, `undislike`.
     * @desc This does validator for `document of like/dislike` and used by post/comment.
     * @param {?} collectionRef
     * @return {?}
     */
    doLikeValidator(collectionRef) {
        console.log(`doLikeValidator() on ${collectionRef.path}`);
        return this.likeDocument(collectionRef).get()
            .then(doc => {
            if (doc.exists) {
                console.log('likeValidator. already liked');
                return this.failure(ALREADY_LIKED); // already liked or disliked.
            }
            else {
                return null; // NOT error. it resolves with null. which means OK.
            }
        })
            .catch(e => {
            // return null;
            console.log(`Caught on validation:Failed to get like/dislike document.This may be a permission error on security rule.`);
            return this.failure(e);
        });
    }
    /**
     * This is a general method for `like`, `unlike`, `dislike`, `disunlike`.
     * @desc The logic is the same for `like` and `dislike` and used by post/comment.
     * @param {?} collectionRef
     * @return {?}
     */
    doLike(collectionRef) {
        console.log(`doLike() on ${collectionRef.path}`);
        return this.doLikeValidator(collectionRef)
            .then(() => {
            console.log(`validator passed. Going to like/dislike on ${collectionRef.path}`);
            return this.likeDocument(collectionRef)
                .set({ time: firestore.FieldValue.serverTimestamp() });
        })
            .then(() => {
            console.log(`like/dislike under ${collectionRef.path} has been added: `);
            return this.countLikes(collectionRef);
        })
            .catch(e => {
            if (e.code === ALREADY_LIKED) {
                console.log(`already like/dislike it. Going to unlike/undislike`);
                return this.doUnlike(collectionRef);
            }
            else {
                console.log(`doLike() failed because: `, e);
                return this.failure(e);
            }
        });
    }
    /**
     *
     * @desc This does unlike/undislike and used by post/comment.
     * @param {?} collectionRef
     * @return {?}
     */
    doUnlike(collectionRef) {
        console.log(`Going to unlike/undislike on ${collectionRef.path}`);
        return this.likeDocument(collectionRef).delete()
            .then(() => {
            console.log(`like/dislike has been deleted(unliked/undisliked) on ${collectionRef.path} `);
            return this.countLikes(collectionRef);
        })
            .catch(e => this.failure(e));
    }
    /**
     * Counts the number of Likes and saves it into `count` document.
     * @desc This cont number of `like/dislike` and used by post/comment.
     * @param {?} collectionRef
     * @return {?}
     */
    countLikes(collectionRef) {
        console.log(`countLikes() on ${collectionRef.path}`);
        return collectionRef.get()
            .then(snapshot => {
            let /** @type {?} */ count = 0;
            if (snapshot.size > 2) {
                // if size is bigger than 2, it probablly has `count` document.
                count = snapshot.size - 1;
            }
            else {
                // if size is 1 or 2, then it may not have `count` document yet.
                snapshot.forEach(doc => {
                    if (doc && doc.exists) {
                        if (doc.id !== 'count') {
                            count++;
                        }
                    }
                });
            }
            console.log(`${collectionRef.path} count: `, count);
            return collectionRef.doc('count').set({ count: count });
        })
            .then(() => {
            console.log(`${collectionRef.path} counted: `);
        });
    }
    /**
     * Returns a promise of <true|false>
     * @desc This is an Action Method.
     * @return {?}
     *      true if the system is already installed.
     *      false if it is not.
     */
    checkInstall() {
        return this.settingsReference.doc('installed').get().then(doc => {
            if (doc && doc.exists) {
                return true;
            }
            else {
                return false;
            }
        })
            .then(re => this.success({ installed: re }))
            .catch(e => this.failure(e));
    }
    /**
     * Installs the system.
     * @desc This is an Action Method.
     * @param {?} options Options to install
     * @return {?}
     */
    install(options) {
        const /** @type {?} */ ref = this.settingsReference.doc('admin');
        console.log(`Going to set admin email on ${ref.path}`);
        return ref.set({ email: options.email }).then(() => {
            const /** @type {?} */ installRef = this.settingsReference.doc('installed');
            console.log(`Admin email is set. Going to set installed.time at ${installRef.path}`);
            return installRef
                .set({ time: firestore.FieldValue.serverTimestamp() });
        })
            .then(re => this.success(true))
            .catch(e => this.failure(e));
    }
}
/**
 * Root collection name.
 */
Base.collectionRoot = COLLECTION_ROOT;
/**
 * Domain collection name under root collection.
 * You can change the collectionDomain programatically.
 * \@code
 * oldDomain
 * constructor(...) {
 * this.oldDomain = Base.collectionDomain;
 * Base.collectionDomain = 'unit-test';
 * }
 * ngOnDestroy() {
 * Base.collectionDomain = this.oldDomain;
 * }
 * \@endcode
 */
Base.collectionDomain = COLLECTION_DOMAIN;
Base.firebase = null;
Base.functions = false;
/**
 * HttpClient object initiated from FireService
 * It is needed to load JSON language files.
 */
Base.http = null;
/**
 *
 * Languages
 * You can change user language with setLanguage()
 */
Base.language = 'en';
Base.languageFolder = 'assets/lang';
Base.texts = { en: en };
/**
 * Settings on how the FireService works
 */
Base.settings = {};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class User extends Base {
    constructor() {
        super(COLLECTIONS.USERS);
    }
    /**
     * @return {?}
     */
    get isLogin() {
        if (this.auth.currentUser) {
            return true;
        }
        else {
            return false;
        }
    }
    /**
     * @return {?}
     */
    get isLogout() {
        return !this.isLogin;
    }
    /**
     * Get current user's uid.
     * @return {?}
     */
    get uid() {
        if (this.auth.currentUser) {
            return this.auth.currentUser.uid;
        }
        else {
            return null;
        }
    }
    /**
     * @return {?}
     */
    get email() {
        if (this.auth.currentUser) {
            return this.auth.currentUser.email;
        }
        else {
            return null;
        }
    }
    /**
     * @return {?}
     */
    get displayName() {
        if (this.auth.currentUser) {
            return this.auth.currentUser.displayName;
        }
        else {
            return null;
        }
    }
    /**
     * @return {?}
     */
    get photoURL() {
        if (this.auth.currentUser) {
            return this.auth.currentUser.photoURL;
        }
        else {
            return null;
        }
    }
    /**
     * Returns user data
     * @return {?}
     */
    data() {
        const /** @type {?} */ user = {
            email: this.email,
            displayName: this.displayName,
            photoURL: this.photoURL,
        };
        return this.collection.doc(this.uid).get().then(doc => {
            if (doc && doc.exists) {
                // console.log('user: ', user, this.auth.currentUser);
                return this.success({ user: Object.assign(user, doc.data()) });
            }
            else {
                return this.failure(USER_NOT_FOUND);
            }
        })
            .catch(e => this.failure(e));
    }
    /**
     * @desc Validates user data to be used for registration
     * @param {?} user User data to validate
     * @return {?}
     */
    registerValidator(user) {
        return __awaiter(this, void 0, void 0, function* () {
            /**PASSWORD VALIDATION**/
            if (!Library.isString(user.password) || !Library.isString(user.email)) {
                return this.failure(FIREBASE_API_ERROR, { info: 'Password and Email should contain valid string.' });
            }
            // As prescribe in `https://stackoverflow.com/questions/98768/should-i-impose-a-maximum-length-on-passwords`
            if (user.password.length > 128) {
                return Promise.reject(PASSWORD_TOO_LONG);
            }
            // if (user.displayName) {
            //     if (user.password.toLowerCase().indexOf(user.displayName.toLowerCase()) > -1) {
            //         return this.failure(WEAK_PASSWORD, { message: 'Password should not contain display name.' });
            //     }
            // }
            // if (user.email) {
            //     const email = user.email.split('@');
            //     if (user.password.toLowerCase().indexOf(email[0].toLowerCase()) > -1) {
            //         return this.failure(WEAK_PASSWORD, { message: 'Password should not contain email.' });
            //     }
            // }
            // if (!user.password.match(/[0-9]/g)) { // must contain number
            //     return this.failure(WEAK_PASSWORD, { message: 'Password should contain atleast 1 number' });
            // }
            // if (!user.password.match(/[a-zA-Z]/g)) { // must contain letter
            //     return this.failure(WEAK_PASSWORD, { message: 'Password should contain atleast 1 letter.' });
            // }
            return null;
        });
    }
    /**
     * User regigration with email and password.
     * @desc this method get user data from HTML FORM including email, password, displayName, gender, birthday, etc.
     * @desc `Firebase Authentication` needs to create an `Authentication` in their `Authentication Service`.
     *      The app
     *          1. Create an Authentication (on the Authentication Service)
     *          2. Update the profile on the Authentication with displayName and photoURL.
     *          3. Sets other information on `users` collection.
     * @param {?} data
     * @return {?}
     */
    register(data) {
        return this.registerValidator(data)
            .then(() => {
            return this.auth.createUserWithEmailAndPassword(data.email, data.password); // 1. create authentication.
        })
            .then((user) => {
            return this.updateAuthentication(user, data); // 2. update Authentication(profile) with `dispalyName` and `photoURL`
        })
            .then((user) => {
            console.log(`Going to set user data under users collection: `);
            return this.set(user, data); // 3. update other information like birthday, gender on `users` collection.
        })
            .then(a => this.success(a))
            .catch(e => this.failure(e));
    }
    /**
     * Validator for User.login()
     * @param {?} email
     * @param {?} password
     * @return {?}
     */
    loginValidator(email, password) {
        return __awaiter(this, void 0, void 0, function* () {
            /**
                    * Test email and password should be both `string`
                    */
            if (!Library.isString(email) || !Library.isString(password)) {
                return this.failure(FIREBASE_API_ERROR, { info: 'Both email and password should contain valid string.' });
            }
            return null;
        });
    }
    /**
     * Login User
     * @param {?} email
     * @param {?} password
     * @return {?}
     */
    login(email, password) {
        return this.loginValidator(email, password)
            .then(() => {
            return this.auth.signInWithEmailAndPassword(email, password);
        })
            .then(a => this.success(a))
            .catch(e => this.failure(e));
        // .catch( e => e );
    }
    /**
     * Logout user.
     * @return {?}
     */
    logout() {
        this.auth.signOut();
    }
    /**
     * Update `displayName`, `photoURL` on Authentication.
     * @desc it does not update other information.
     * @see `this.updateProfile()` for updating user information.
     * @param {?} user
     * @param {?} data
     * @return {?}
     */
    updateAuthentication(user, data) {
        const /** @type {?} */ up = {
            displayName: data.displayName,
            photoURL: data.photoURL
        };
        console.log('updateAuthentication: up: ', up);
        return user.updateProfile(Library.sanitize(up)).then(x => user);
    }
    /**
     * Sets user information on user collection.
     * @param {?} user
     * @param {?} data
     * @return {?}
     */
    set(user, data) {
        delete data.displayName;
        delete data.photoURL;
        delete data.password;
        data.created = firestore.FieldValue.serverTimestamp();
        // return this.collection.doc(user.uid).set(data).then(x => user);
        data.uid = this.uid;
        return this.create(data)
            .catch(e => {
            console.log(`Failed to set data under users collection: `);
            return this.failure(e);
        });
    }
    /**
     * @param {?} data
     * @return {?}
     */
    create(data) {
        const /** @type {?} */ ref = this.collection.doc(data.uid);
        console.log(`set at: ${ref.path} with: `, data);
        return ref.set(data)
            .then(() => this.success({ id: data.uid }))
            .catch(e => this.failure(e));
    }
    /**
     * @param {?} data
     * @return {?}
     */
    update(data) {
        data.uid = this.uid;
        console.log('update: ', data);
        return this.collection.doc(data.uid).set(data)
            .then(() => this.success({ id: data.uid }))
            .catch(e => this.failure(e));
    }
    /**
     * @return {?}
     */
    delete() {
        console.log('delete: ', this.uid);
        return this.collection.doc(this.uid).delete()
            .then(() => this.success(null))
            .catch(e => this.failure(e));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class Category extends Base {
    constructor() {
        super(COLLECTIONS.CATEGORIES);
    }
    /**
     * Creates a category
     *
     * @param {?} category
     * @return {?}
     *      - Promise<CATEGORY_CREATE>
     *      - Otherwise Promise<RESONSE> error object.
     */
    createValidator(category) {
        const /** @type {?} */ idCheck = this.checkDocumentIDFormat(category.id);
        if (idCheck) {
            return this.failure(new Error(idCheck), { documentID: category.id });
        }
        return this.collection.doc(category.id).get()
            .then(doc => {
            if (doc.exists) {
                return this.failure(new Error(CATEGORY_EXISTS));
            }
            else {
                return null; // NOT error.
            }
        })
            .catch(e => this.failure(e)); // NOT error. The error is handled here and resolves with null.
    }
    /**
     * @param {?} category
     * @return {?}
     */
    create(category) {
        return this.createValidator(category)
            .then(() => {
            category.subcategories = Library.removeSpaceBetween(',', category.subcategories);
            category.created = firestore.FieldValue.serverTimestamp();
            const /** @type {?} */ ref = this.collection.doc(category.id);
            console.log(`set at: ${ref.path}`);
            return ref.set(Library.sanitize(category));
        })
            .then(() => this.success({ id: category.id }))
            .catch(e => {
            return this.failure(e);
        });
        // const validate = this.createValidator(category);
        // if (validate) {
        //     return validate;
        // }
        // category.subcategories = _.removeSpaceBetween(',', category.subcategories);
        // category.created = firebase.firestore.FieldValue.serverTimestamp();
        // return this.collection.doc(category.id).get()
        //     .then(doc => {
        //         if (doc.exists) {
        //             return this.failure(new Error(CATEGORY_EXISTS));
        //         } else {
        //             return this.collection.doc(category.id).set(_.sanitize(category));
        //         }
        //     })
        //     .then(() => {
        //         return this.success(category.id);
        //     })
        //     .catch(e => this.failure(e));
    }
    /**
     * Edits a category.
     *
     * \@todo category exists validation.
     * @param {?} category
     * @return {?}
     */
    editValidator(category) {
        return __awaiter(this, void 0, void 0, function* () {
            const /** @type {?} */ idCheck = this.checkDocumentIDFormat(category.id);
            if (idCheck) {
                return this.failure(new Error(idCheck), { documentID: category.id });
            }
            return null;
        });
    }
    /**
     * @param {?} category
     * @return {?}
     */
    edit(category) {
        return this.editValidator(category)
            .then(() => this.collection.doc(category.id).update(Library.sanitize(category)))
            .then(() => this.success({ id: category.id }))
            .catch(e => this.failure(e));
        // category.subcategories = _.removeSpaceBetween(',', category.subcategories);
        // return this.collection.doc(category.id).update(_.sanitize(category))
        //     .then(() => this.success(category.id))
        //     .catch(e => this.failure(e));
    }
    /**
     *
     * @param {?} id
     * @return {?}
     */
    deleteValidator(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const /** @type {?} */ idCheck = this.checkDocumentIDFormat(id);
            if (idCheck) {
                return this.failure(new Error(idCheck), { documentID: id });
            }
            return yield this.collection.doc(id).get()
                .then(doc => {
                if (doc.exists) {
                    const /** @type {?} */ data = /** @type {?} */ (doc.data());
                    if (data.numberOfPosts) {
                        return this.failure(COLLECTION_NOT_EMPTY);
                    }
                    else {
                        return null;
                    }
                }
                else {
                    return this.failure(new Error(NOT_FOUND));
                }
            });
        });
    }
    /**
     * @param {?} id
     * @return {?}
     */
    delete(id) {
        return this.deleteValidator(id)
            .then(() => {
            return this.collection.doc(id).delete();
        })
            .catch(e => this.failure(e));
    }
    /**
     * @return {?}
     */
    categories() {
        return this.collection.get().then((querySnapshot) => {
            const /** @type {?} */ categories = /** @type {?} */ ([]);
            querySnapshot.forEach(doc => {
                categories.push(/** @type {?} */ (doc.data()));
            });
            return categories;
        });
    }
    /**
     * @param {?} id
     * @return {?}
     */
    get(id) {
        return super.get(id);
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class Post extends Base {
    constructor() {
        super(COLLECTIONS.POSTS); // Base constructor to Base.collectionName
        /**
         * Navigation
         * `cursor` is indicating where to load from. It is null by default.
         */
        this.cursor = null;
        /**
         * Navigation
         * `categoryId` holds the selected category id to load and display posts from.
         * null by default
         */
        this.categoryId = null;
        /**
         * Posts and its IDs that have been loaded by `page()`.
         * Since object has no sequence, `pagePostIds` is holding the keys of the posts in order.
         * These are public variasble which should be used on list component to display posts.
         */
        this.pagePosts = {};
        this.pagePostIds = [];
        /**
         * Subscribing changes for realtime update.
         */
        this._unsubscribeLikes = [];
        this._unsubscribePosts = [];
        this.unsubscribePage = null;
        this.user = new User();
    }
    /**
     * Validates the input data for creating a post.
     * @desc validate the input for creating a post.
     * @desc Don't check if the category id is really exists. Normally this won't make a trouble.
     * @param {?} post
     * @return {?}
     */
    createValidator(post) {
        if (this.user.isLogout) {
            return Promise.reject(new Error(USER_IS_NOT_LOGGED_IN));
        }
        if (post.id) {
            return Promise.reject(new Error(POST_ID_NOT_EMPTY));
        }
        if (Library.isEmpty(post.category)) {
            return Promise.reject(new Error(CATEGORY_ID_EMPTY));
        }
        // if ( !_.isEqual(post.uid, this.user.uid) ) {
        //     return this.failure(PERMISSION_DENIED, {info: 'You cannot post on behalf of other users.'});
        // }
        return Promise.resolve(null);
    }
    /**
     * @param {?} post
     * @return {?}
     */
    createSanitizer(post) {
        Library.sanitize(post);
        post.uid = this.user.uid;
        post.date = this.user.displayName;
        post.created = firestore.FieldValue.serverTimestamp();
        // console.log(post);
        return post;
    }
    /**
     * Creates a post.
     * @desc if `post.id` exists, then it rejects.
     * \@since 2018-03-16 Category.numberOfPosts were removed. \@see README## Client Side Coding Limitation and PUBLIC META DATA
     * @param {?} post
     * @return {?} Promise<POST_CREATE> with Document ID if success.
     */
    create(post) {
        return this.createValidator(post)
            .then(() => {
            return this.collection.add(this.createSanitizer(post));
            // const categoryRef = this.db.collection(COLLECTIONS.CATEGORIES).doc(post.category);
            // const postRef = this.db.collection(COLLECTIONS.POSTS).doc();
            // return <any>this.db.runTransaction(t => {
            //     return t.get(categoryRef)
            //         .then(category => {
            //             if (!category.exists) {
            //                 throw CATEGORY_DOES_NOT_EXIST;
            //             }
            //             const categoryData = <CATEGORY>category.data();
            //             if (categoryData.numberOfPosts === void 0) {
            //                 categoryData.numberOfPosts = 1;
            //             } else {
            //                 categoryData.numberOfPosts++;
            //             }
            //             t
            //                 .set(categoryRef, categoryData)
            //                 .set(postRef, post);
            //             return postRef.id;
            //         });
            // });
        })
            .then(doc => {
            return this.success({ id: doc.id, post: post });
        })
            .catch(e => this.failure(e));
    }
    /**
     * Validates edit.
     * @desc on edit, category can be empty.
     * @param {?} post
     * @return {?}
     */
    editValidator(post) {
        if (this.user.isLogout) {
            return Promise.reject(new Error(USER_IS_NOT_LOGGED_IN));
        }
        if (Library.isEmpty(post.id)) {
            return Promise.reject(new Error(POST_ID_EMPTY));
        }
        // if (_.isEmpty(post.category)) {
        //     return Promise.reject(new Error(CATEGORY_ID_EMPTY));
        // }
        return Promise.resolve(null);
    }
    /**
     * @param {?} post
     * @return {?}
     */
    edit(post) {
        return /** @type {?} */ (this.editValidator(post)
            .then(() => {
            Library.sanitize(post);
            post.updated = firestore.FieldValue.serverTimestamp();
            const /** @type {?} */ ref = this.collection.doc(post.id);
            console.log('update at: ', ref.path);
            return ref.update(post);
        })
            .then(() => {
            return this.success({ id: post.id, post: post });
        })
            .catch(e => this.failure(e)));
    }
    /**
     * Deletes a post.
     *
     * @desc It puts the post id under `posts_deleted` to indicate that the post has been deleted.
     * @see REAMEMD## posts_deleted collection
     * \@todo test on deleting and marking.
     * @param {?} id
     * @return {?}
     */
    delete(id) {
        const /** @type {?} */ post = {
            id: id,
            title: POST_DELETED,
            content: POST_DELETED,
            deleted: true
        };
        return this.edit(post);
        // return this.collection.doc(id).delete()
        //     .then(() => {
        //         return this.db.collection(COLLECTIONS.POSTS_DELETED).doc(id)
        //             .set({ time: firebase.firestore.FieldValue.serverTimestamp() });
        //     })
        //     .then(() => this.success({ id: id }))
        //     .catch(e => this.failure(e));
    }
    /**
     * It remembers previous category for pagnation.
     * If category changes, it will clear the cursor.
     * @param {?} category
     * @return {?}
     */
    categoryChanged(category) {
        return this.categoryId !== category;
    }
    /**
     * For pagination.
     *
     * @param {?} category
     * @return {?}
     */
    resetCursor(category) {
        this.categoryId = category;
        this.cursor = null;
    }
    /**
     * Get posts for a page.
     * @desc if input `category` is given, then it opens a new category and gets posts for the first page.
     *    Otherwise it gets posts for next page.
     *
     * @param {?} category
     * @return {?} true if the category has been chagned and reset. other wise false.
     */
    resetLoadPage(category) {
        let /** @type {?} */ reset = false;
        if (category) {
            if (category === 'all' || this.categoryId !== category) {
                this.pagePosts = {};
                this.pagePostIds = [];
                this.categoryId = category;
                this.resetCursor(category);
                this.unsubscribePosts();
                this.unsubscribeLikes();
                reset = true;
            }
        }
        return reset;
    }
    /**
     * Unsubscribe all the posts.
     * @return {?}
     */
    unsubscribePosts() {
        if (this._unsubscribePosts.length) {
            this._unsubscribePosts.map(unsubscribe => unsubscribe());
        }
    }
    /**
     * @return {?}
     */
    unsubscribeLikes() {
        if (this._unsubscribeLikes.length) {
            this._unsubscribeLikes.map(unsubscribe => {
                unsubscribe();
            });
        }
    }
    /**
     * @param {?} post
     * @return {?}
     */
    subscribePostChange(post) {
        if (!this.settings.listenOnPostChange) {
            return;
        }
        const /** @type {?} */ path = this.post(post.id).path;
        const /** @type {?} */ unsubscribe = this.post(post.id).onSnapshot(doc => {
            // console.log('Update on :', path, doc.data());
            post = Object.assign(post, doc.data());
        });
        this._unsubscribePosts.push(unsubscribe);
    }
    /**
     * Subscribes for likes/dislikes
     * @param {?} post post to subscribe for like, dislike
     * @return {?}
     */
    subscribeLikes(post) {
        if (!this.settings.listenOnPostLikes) {
            return;
        }
        const /** @type {?} */ likeRef = this.likeColllection(post.id, COLLECTIONS.LIKES).doc('count');
        // console.log('subscribe on likes: ', post.id, `path: ${likeRef.path}`);
        const /** @type {?} */ subscribeLik = likeRef.onSnapshot(doc => {
            if (doc.exists) {
                const /** @type {?} */ data = doc.data();
                post.numberOfLikes = data["count"];
            }
        });
        this._unsubscribeLikes.push(subscribeLik);
        const /** @type {?} */ dislikeRef = this.likeColllection(post.id, COLLECTIONS.DISLIKES).doc('count');
        // console.log('subscribe on dislikes: ', post.id, `path: ${dislikeRef.path}`);
        const /** @type {?} */ subscribeDislike = dislikeRef.onSnapshot(doc => {
            // console.log('changed on dislike: ', doc);
            if (doc.exists) {
                const /** @type {?} */ data = doc.data();
                post.numberOfDislikes = data["count"];
            }
        });
        this._unsubscribeLikes.push(subscribeDislike);
    }
    /**
     * Gets post data to display on the page.
     *
     * 1. Build the `query` for parsing data to firebase.
     *  `collection()` - post collection super(COLLECTIONS.POST)
     *  `.where()` - category
     *  `.orderBy()` - created, desc
     *  `.startAfter()` - last doc queried
     *  `.limit()` - limits doc tobe queried.
     *
     * - Will get data based on query above.
     *  For each document `doc` parsed. We do.
     * - Gets document as object. as what `doc.data()` does.
     * - Pushes additional properties `id` amd `date` post date created.
     * - Subscribe or listen to `post changes` and `likes`.
     *
     * - Updates `cursor` based on the last `document` parsed.
     *
     * - if category changes `Post.page()` will start again and unsubscribe to posts from previous category.
     *
     * @param {?} options
     * @return {?} Promise<Array<POST>> posts that are loaded.
     *
     *
     */
    page(options) {
        const /** @type {?} */ reset = this.resetLoadPage(options.category);
        let /** @type {?} */ query = /** @type {?} */ (this.collection);
        if (this.categoryId && this.categoryId !== 'all') {
            query = query.where('category', '==', this.categoryId);
        }
        query = query.orderBy('created', 'desc');
        if (this.cursor) {
            query = query.startAfter(this.cursor);
        }
        query = query.limit(options.limit);
        return /** @type {?} */ (query.get().then(querySnapshot => {
            if (querySnapshot.docs.length) {
                querySnapshot.forEach(doc => {
                    const /** @type {?} */ post = /** @type {?} */ (doc.data());
                    post.id = doc.id;
                    post['date'] = (new Date(post.created)).toLocaleString();
                    this.pagePosts[post.id] = post;
                    this.pagePostIds.push(post.id);
                    this.subscribePostChange(post);
                    this.subscribeLikes(post);
                });
                // only one cursor is supported and normally one page has on pagination.
                this.cursor = querySnapshot.docs[querySnapshot.docs.length - 1];
                // @see comment on subscribeNewPost()
                if (reset) {
                    this.subscribePostAdd(query);
                }
                return this.pagePosts;
            }
            else {
                return [];
            }
        }));
    }
    /**
     * Listens on new post only. It does not listen for edit/delete.
     *  It subscribes added/updated/removed only after loading/displaying the post list.
     *  In this way, it prevents double display of the last post.
     * @param {?} query
     * @return {?}
     */
    subscribePostAdd(query) {
        if (!this.settings.listenOnPostChange) {
            return;
        }
        if (this.unsubscribePage) {
            this.unsubscribePage();
        }
        this.unsubscribePage = query.limit(1).onSnapshot(snapshot => {
            snapshot.docChanges.forEach(change => {
                const /** @type {?} */ doc = change.doc;
                if (doc.metadata.hasPendingWrites) {
                    console.log('pending', doc.metadata.hasPendingWrites, 'from cache: ', doc.metadata.fromCache);
                }
                else {
                    console.log('pending', doc.metadata.hasPendingWrites, 'type: ', change.type, 'from cache: ', doc.metadata.fromCache, doc.data());
                    const /** @type {?} */ post = doc.data();
                    post.id = doc.id;
                    console.log(`exists: ${this.pagePosts[post.id]}`);
                    if (change.type === 'added' && this.pagePosts[post.id] === void 0) {
                        this.addPostOnTop(post);
                    }
                    else if (change.type === 'modified') {
                        this.updatePost(post);
                    }
                    else if (change.type === 'removed') {
                        this.removePost(post);
                    }
                }
            });
        });
    }
    /**
     * Add a newly created post on top of post list on the page
     *  - and subscribe post changes if `settings.listenPostChange` is set to true.
     *  - and subscribe like/dislike based on the settings.
     *
     * @desc It's important to understand how `added` event fired on `onSnapshot)`.
     *
     * @param {?} post
     * @return {?}
     */
    addPostOnTop(post) {
        if (this.pagePosts[post.id] === void 0) {
            console.log(`addPostOnTop: `, post);
            this.pagePosts[post.id] = post;
            this.pagePostIds.unshift(post.id);
            this.subscribePostChange(post);
            this.subscribeLikes(post);
        }
    }
    /**
     * When listening the last post on collection in realtime, it often fires `modified` event on new docuemnt created.
     * @param {?} post
     * @return {?}
     */
    updatePost(post) {
        console.log('updatePost id: ', post.id);
        if (this.pagePosts[post.id]) {
            console.log(`updatePost`, post);
            this.pagePosts[post.id] = Object.assign(this.pagePosts[post.id], post);
        }
        else {
            this.addPostOnTop(post);
        }
    }
    /**
     * This method is no longer in use.
     *
     * @deprecated \@see README### No post delete.
     * @param {?} post
     * @return {?}
     */
    removePost(post) {
        // if (this.pagePosts[post.id]) {
        //     console.log(`deletePost`, post);
        //     this.pagePosts[post.id].title = 'deleted???';
        //     // delete this.pagePosts[post.id];
        // }
    }
    /**
     * @return {?}
     */
    stopLoadPage() {
        this.resetLoadPage(undefined);
        this.unsubscribeLikes();
        this.unsubscribePosts();
    }
    /**
     * Returns post docuement reference.
     * @param {?} postId
     * @return {?}
     */
    post(postId) {
        return this.collection.doc(postId);
    }
    /**
     * Returns collection of like/dislike.
     * @param {?} postId Post Document ID
     * @param {?} collectionName Subcollection name
     * @return {?}
     */
    likeColllection(postId, collectionName) {
        return this.collection.doc(postId)
            .collection(collectionName);
    }
    /**
     * @param {?} postId
     * @return {?}
     */
    like(postId) {
        return this.doLike(this.likeColllection(postId, COLLECTIONS.LIKES));
    }
    /**
     * @param {?} postId
     * @return {?}
     */
    dislike(postId) {
        return this.doLike(this.likeColllection(postId, COLLECTIONS.DISLIKES));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class Data extends Base {
    constructor() {
        super('');
        this.user = new User();
    }
    /**
     * Get the storage data reference of a file.
     * @param {?} file
     * @return {?}
     */
    getDataRef(file) {
        return storage().ref().child(`${this.user.uid}`).child(`${file.name}`);
    }
    /**
     * @param {?} data
     * @return {?}
     */
    delete(data) {
        return storage().ref(data.fullPath).delete().then(re => {
            return this.success(true);
        })
            .catch(e => this.failure(e));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class Comment extends Base {
    constructor() {
        super(COLLECTIONS.COMMENTS);
        this.comments = {};
        this.commentIds = {};
        /**
         * Subscribes and unsubscribes comment's likes/dislike/changes by post.
         * This is because in some cases, only few posts may be destroyed from page and
         * it detaches the listeners of those posts only.
         */
        this._unsubscribeLikes = {};
        this._unsubscribeComments = {};
        this.user = new User();
    }
    /**
     * Returns parent comment.
     * @param {?} commentId
     * @return {?}
     */
    getParentComment(commentId) {
        if (commentId) {
            const /** @type {?} */ comment = this.getComment(commentId);
            if (comment.parentId) {
                return this.getComment(comment.parentId);
            }
            else {
                return /** @type {?} */ ([]);
            }
        }
        return /** @type {?} */ ([]);
    }
    /**
     * Returns the comment.
     * @param {?} commentId
     * @return {?}
     */
    getComment(commentId) {
        if (this.comments[commentId] === void 0) {
            return /** @type {?} */ ([]);
        }
        // console.log(this.fire.comment.comments[id]);
        return this.comments[commentId];
    }
    /**
     * Comment Collection under a post.
     * @param {?} postId the post document id.
     * @return {?}
     */
    commentCollection(postId) {
        return this.collectionRef(COLLECTIONS.POSTS).doc(postId).collection(COLLECTIONS.COMMENTS);
    }
    /**
     * Returns comment reference.
     * @param {?} postId
     * @param {?} commentId
     * @return {?}
     */
    comment(postId, commentId) {
        return this.commentCollection(postId).doc(commentId);
    }
    /**
     * Loads all the comments under the postId.
     * @param {?} postId
     * @return {?}
     */
    load(postId) {
        const /** @type {?} */ ref = this.commentCollection(postId);
        // console.log(`gets at: ${ref.path}`);
        return ref.orderBy('created', 'asc').get().then(s => {
            s.forEach(doc => {
                const /** @type {?} */ c = doc.data();
                c.id = doc.id;
                this.comments[c.id] = c;
                // @todo sort comments by thread.
                const /** @type {?} */ sorted = this.sortComments(postId, c);
                this.subscribeCommentChange(postId, c);
                this.subscribeLikes(c);
            });
            // console.log('sorted: ', sorted);
            // observe like/dislike
            this.subscribeCommentAdd(postId);
            return this.commentIds[postId];
        }).catch(e => this.failure(e));
    }
    /**
     * Sorts the comments.
     * @desc The app may need to rerender the page depending on the app context.
     * @param {?} postId Post Document ID
     * @param {?} comment
     * @return {?}
     */
    sortComments(postId, comment) {
        console.log(`sortComment for: `, postId, comment);
        if (this.commentIds[postId] === void 0) {
            this.commentIds[postId] = [];
        }
        const /** @type {?} */ pos = this.commentIds[postId].findIndex(id => id === comment.parentId);
        console.log('pos: ', pos);
        if (pos === -1) {
            this.commentIds[postId].push(comment.id);
        }
        else {
            this.commentIds[postId].splice(pos + 1, 0, comment.id);
        }
        // this.commentIds[postId].push(comment.id);
        return this.commentIds[postId];
    }
    /**
     * @param {?} comment
     *      comment[postId] and comment[parentId] is already set.
     * @return {?}
     */
    create(comment) {
        Library.sanitize(comment);
        comment.uid = this.user.uid;
        comment.displayName = this.user.displayName;
        comment.created = firestore.FieldValue.serverTimestamp();
        if (comment.parentId) {
            comment.depth = this.getComment(comment.parentId).depth + 1;
        }
        else {
            comment.depth = 0;
        }
        const /** @type {?} */ ref = this.commentCollection(comment.postId);
        console.log(`Going to add a comment under: ${ref.path}`);
        return ref.add(comment)
            .then(doc => {
            return this.success({ id: doc.id });
        })
            .catch(e => {
            console.log(`failed: `, e);
            this.failure(e);
        });
    }
    /**
     * @param {?} comment
     * @return {?}
     */
    edit(comment) {
        Library.sanitize(comment);
        comment.uid = this.user.uid;
        comment.postId = this.comments[comment.id].postId;
        comment.updated = firestore.FieldValue.serverTimestamp();
        const /** @type {?} */ ref = this.comment(comment.postId, comment.id);
        console.log(`Going to edit : ${ref.path} with `, comment);
        return ref.update(comment).then(() => {
            return this.success({ id: comment.id });
        })
            .catch(e => this.failure(e));
    }
    /**
     * Returns collection of comment like/dislike.
     * @param {?} commentId Comment Document ID
     * @param {?} collectionName Subcollection name under comment
     * @return {?}
     */
    likeColllection(commentId, collectionName) {
        const /** @type {?} */ postId = this.comments[commentId].postId;
        return this.comment(postId, commentId)
            .collection(collectionName);
    }
    /**
     * @param {?} commentId
     * @return {?}
     */
    like(commentId) {
        return this.doLike(this.likeColllection(commentId, COLLECTIONS.LIKES));
    }
    /**
     * @param {?} commentId
     * @return {?}
     */
    dislike(commentId) {
        return this.doLike(this.likeColllection(commentId, COLLECTIONS.DISLIKES));
    }
    /**
     *
     * @param {?} postId
     * @param {?} comment
     * @return {?}
     */
    subscribeCommentChange(postId, comment) {
        if (!this.settings.listenOnCommentChange) {
            return;
        }
        const /** @type {?} */ unsubscribe = this.commentCollection(postId).doc(comment.id).onSnapshot(doc => {
            comment = Object.assign(comment, doc.data());
        });
        this.pushCommentChangeSubscriber(postId, unsubscribe);
    }
    /**
     * @param {?} postId
     * @return {?}
     */
    subscribeCommentAdd(postId) {
        if (!this.settings.listenOnCommentChange) {
            return;
        }
        const /** @type {?} */ path = this.commentCollection(postId).path;
        console.log(`watch for new comment: ${path}`);
        const /** @type {?} */ unsubscribe = this.commentCollection(postId).orderBy('created', 'desc').limit(1).onSnapshot(snapshot => {
            console.log('watch the lastest comment: ');
            snapshot.docChanges.forEach(change => {
                const /** @type {?} */ doc = change.doc;
                if (doc.metadata.hasPendingWrites) {
                    console.log('pending', doc.metadata.hasPendingWrites, 'from cache: ', doc.metadata.fromCache);
                }
                else {
                    console.log('pending', doc.metadata.hasPendingWrites, 'type: ', change.type, 'from cache: ', doc.metadata.fromCache, doc.data());
                    const /** @type {?} */ comment = doc.data();
                    comment.id = doc.id;
                    // console.log(`exists: ${this.pagePosts[post.id]}`);
                    if (change.type === 'added' && this.comments[comment.id] === void 0) {
                        this.insertComment(postId, comment);
                    }
                    else if (change.type === 'modified') {
                        this.updateComment(postId, comment);
                    }
                    else if (change.type === 'removed') {
                        this.removeComment(comment);
                    }
                }
            });
        });
        this.pushCommentChangeSubscriber(postId, unsubscribe);
    }
    /**
     * @param {?} postId
     * @param {?} unsubscribe
     * @return {?}
     */
    pushCommentChangeSubscriber(postId, unsubscribe) {
        if (this._unsubscribeComments[postId] === void 0) {
            this._unsubscribeComments[postId] = [];
        }
        this._unsubscribeComments[postId].push(unsubscribe);
    }
    /**
     * Add a newly created post on top of post list on the page
     *  - and subscribe post changes if `settings.listenPostChange` is set to true.
     *  - and subscribe like/dislike based on the settings.
     *
     * @desc It's important to understand how `added` event fired on `onSnapshot)`.
     *
     * @param {?} postId
     * @param {?} comment
     * @return {?}
     */
    insertComment(postId, comment) {
        if (this.comments[comment.id] === void 0) {
            console.log(`insertComment: `, comment);
            this.comments[comment.id] = comment;
            this.sortComments(postId, comment);
            this.subscribeCommentChange(postId, comment);
            this.subscribeLikes(comment);
        }
    }
    /**
     * When listening the last post on collection in realtime, it often fires `modified` event on new docuemnt created.
     * @param {?} postId
     * @param {?} comment
     * @return {?}
     */
    updateComment(postId, comment) {
        console.log('updateComment id: ', comment.id);
        if (this.comments[comment.id]) {
            console.log(`updateComment`, comment);
            this.comments[comment.id] = Object.assign(this.comments[comment.id], comment);
        }
        else {
            this.insertComment(postId, comment);
        }
    }
    /**
     * This method is no longer in use.
     *
     * @deprecated \@see README### No post delete.
     * @param {?} comment
     * @return {?}
     */
    removeComment(comment) {
    }
    /**
     * @param {?} comment
     * @return {?}
     */
    subscribeLikes(comment) {
        if (!this.settings.listenOnCommentLikes) {
            return;
        }
        const /** @type {?} */ likeRef = this.likeColllection(comment.id, COLLECTIONS.LIKES).doc('count');
        // console.log('subscribe on likes: ', post.id, `path: ${likeRef.path}`);
        const /** @type {?} */ subscribeLike = likeRef.onSnapshot(doc => {
            if (doc.exists) {
                const /** @type {?} */ data = doc.data();
                comment.numberOfLikes = data["count"];
            }
        });
        // this._unsubscribeLikes.push(subscribeLik);
        this.pushCommentLikeSubscriber(comment.postId, subscribeLike);
        const /** @type {?} */ dislikeRef = this.likeColllection(comment.id, COLLECTIONS.DISLIKES).doc('count');
        // console.log('subscribe on dislikes: ', post.id, `path: ${dislikeRef.path}`);
        const /** @type {?} */ subscribeDislike = dislikeRef.onSnapshot(doc => {
            // console.log('changed on dislike: ', doc);
            if (doc.exists) {
                const /** @type {?} */ data = doc.data();
                comment.numberOfDislikes = data["count"];
            }
        });
        // this._unsubscribeLikes.push(subscribeDislike);
        this.pushCommentLikeSubscriber(comment.postId, subscribeDislike);
    }
    /**
     * @param {?} postId
     * @param {?} unsubscribe
     * @return {?}
     */
    pushCommentLikeSubscriber(postId, unsubscribe) {
        if (this._unsubscribeLikes[postId] === void 0) {
            this._unsubscribeLikes[postId] = [];
        }
        this._unsubscribeLikes[postId].push(unsubscribe);
    }
    /**
     * Unsubscribe for the realtime update of changes/likes/dislikes.
     * If you don't unsubscribe, you have to pay more since it will still listen(read) the documents that are no longer needed.
     * @param {?} postId
     * @return {?}
     */
    unsubscribes(postId) {
        if (this._unsubscribeComments[postId] !== void 0 && this._unsubscribeComments[postId].length) {
            this._unsubscribeComments[postId].map(unsubscribe => unsubscribe());
        }
        if (this._unsubscribeLikes[postId] !== void 0 && this._unsubscribeLikes[postId].length) {
            this._unsubscribeLikes[postId].map(unsubscribe => unsubscribe());
        }
    }
    /**
     * Destroys all the resources that were used to display comments of a post.
     * @desc it should be called from `OnDestroy` of the comment component
     *      or anywhere you want to destroy the comments that belong to the post.
     * @param {?} post
     * @return {?}
     */
    destory(post) {
        console.log(`Going to destroy comments for post.id: `, post.id);
        this.unsubscribes(post.id);
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class FireService extends Base {
    /**
     * This runs only one time. contructor of Service will run only one time and re-used by container.
     * @param {?} ngZone
     * @param {?} http
     */
    constructor(ngZone, http) {
        super();
        this.ngZone = ngZone;
        this.user = new User();
        this.category = new Category();
        this.post = new Post();
        this.comment = new Comment();
        this.data = new Data();
        Base.http = http;
        this.initUser();
    }
    /**
     * @return {?}
     */
    initUser() {
        auth().onAuthStateChanged((user) => {
            this.ngZone.run(x => { });
            
        });
    }
}
FireService.decorators = [
    { type: Injectable },
];
/** @nocollapse */
FireService.ctorParameters = () => [
    { type: NgZone, },
    { type: HttpClient, },
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class FirelibraryModule {
    /**
     * @param {?} config
     * @return {?}
     */
    static forRoot(config) {
        Base.configure(config);
        return {
            ngModule: FirelibraryModule,
            providers: [FireService],
        };
    }
}
FirelibraryModule.decorators = [
    { type: NgModule, args: [{
                imports: [
                    CommonModule,
                    HttpClientModule
                ],
                declarations: [],
                providers: [FireService]
            },] },
];
/** @nocollapse */
FirelibraryModule.ctorParameters = () => [];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * Generated bundle index. Do not edit.
 */

export { FirelibraryModule, FireService, Library as _, Base, COLLECTIONS, POST_DELETED, UNKNOWN, FIREBASE_API_ERROR, NOT_FOUND, ALREADY_EXISTS, PERMISSION_DENIED, RESOURCE_EXHAUSTED, USER_IS_NOT_LOGGED_IN, NO_DOCUMENT_ID, DOCUMENT_ID_TOO_LONG, DOCUMENT_ID_CANNOT_CONTAIN_SLASH, COLLECTION_NOT_EMPTY, CATEGORY_ID_EMPTY, CATEGORY_EXISTS, CATEGORY_DOES_NOT_EXIST, POST_ID_NOT_EMPTY, POST_ID_EMPTY, ALREADY_LIKED, EMAIL_ALREADY_IN_USE, USER_NOT_FOUND, INVALID_EMAIL, WEAK_PASSWORD, EXPIRED_ID_TOKEN, WRONG_PASSWORD, PASSWORD_TOO_LONG };
//# sourceMappingURL=firelibrary.js.map
