import { HttpClient } from '@angular/common/http';
import * as firebase from 'firebase';
import { SYSTEM_CONFIG, INSTALLED, FIRESERVICE_SETTINGS } from './interface';
export * from './interface';
export { Library as _ } from './library';
export * from './define';
export * from './error';
export declare class Base {
    collectionName: string;
    /**
     * Root collection name.
     */
    static collectionRoot: string;
    /**
     * Domain collection name under root collection.
     * You can change the collectionDomain programatically.
     * @code
            oldDomain
            constructor(...) {
                this.oldDomain = Base.collectionDomain;
                Base.collectionDomain = 'unit-test';
            }
            ngOnDestroy() {
                Base.collectionDomain = this.oldDomain;
            }
     * @endcode
     */
    static collectionDomain: string;
    static firebase: firebase.app.App;
    static functions: boolean;
    /**
     * HttpClient object initiated from FireService
     * It is needed to load JSON language files.
     */
    static http: HttpClient;
    /**
    *
    * Languages
    * You can change user language with setLanguage()
    */
    static language: string;
    static languageFolder: string;
    static texts: {
        [language: string]: any;
    };
    /**
     * Settings on how the FireService works
     */
    static settings: FIRESERVICE_SETTINGS;
    auth: firebase.auth.Auth;
    db: firebase.firestore.Firestore;
    storage: firebase.storage.Storage;
    constructor(collectionName?: string);
    static configure(config: SYSTEM_CONFIG): void;
    version(): string;
    readonly settings: FIRESERVICE_SETTINGS;
    readonly collectionRoot: string;
    readonly collectionDomain: string;
    setSettings(obj: FIRESERVICE_SETTINGS): void;
    readonly http: HttpClient;
    /**
    * Returns the collection of the selected collection in this.collectionName
    */
    readonly collection: firebase.firestore.CollectionReference;
    /**
     * Returns a collection reference under the domain.
     * @desc do not use this.db.collection(....)
     *          Use this method to get a collection reference.
     * @returns a collection reference.
     *
     */
    collectionRef(name: any): firebase.firestore.CollectionReference;
    readonly settingsReference: firebase.firestore.CollectionReference;
    /**
    *
    * Returns an object of RESPONSE interface.
    *
    * @returns
    *      - <RESPONE> object.
    */
    success(data?: any): any;
    /**
    *
    * Returns a Promise.
    *
    * @desc This method returns a Promise. And this is important to understand.
    *      Firstly, it will first return a Promise.
    *      Secondl, laster it will reject().
    *
    * @desc This method is actually an Asynchronous method since it returns a Promise.
    *
    * @desc `_count` counts how many times the `Error Object` passed over to this method.
    *          - when an Error that has already done translate with `failture()`, the does not translate again.
    *              it simply increase `_count`.
    *              so, it is safe to pass one object to this method over gain.
    *
    *
    * @param obj error object or error code string.
    *      If the input `obj` is a string, then it encapsulates with ` new Error( obj ) `.
    *      So, the input `obj` is transformed into an Error object.
    *      If the input `obj` is an Error object, It may be a Firebase Error.
    *      It will detect if it's a Firebase Error and translates the Firebase Error into proper `error message`.
    *          ( The error code of Firebase Error should not be changed. Only error message should be translated. )
    *      Unfortunately we cannot use instance of `firebase.firestore.FirestoreError` or we do not know how to use it yet.
    *      So, we simply compare if the error code is one of the firebase error code.
    *      @see for Firestore, https://firebase.google.com/docs/reference/js/firebase.firestore.FirestoreError
    *      @see for Authentication, https://firebase.google.com/docs/reference/js/firebase.auth.Error
    *
    * @return Promise<never>
    *
    * @code this.failure( UNKNOWN );
    * @code this.failure( new Error(UNKNOWN) ); // same as above.
    *
    * @example test/test.error.ts
    * @see test.error.ts on how it throws error.
    *
    */
    failure(obj: Error | string, info?: {}): Promise<never>;
    /**
    * Returns true if the error is a Firebase Error object.
    *
    * @description
    *      - It checks if the error code is the same as `Firebase Error Code`. If yes, it returns true.
    */
    isFirebaseError(e: any, info: any): boolean;
    translateFirebaseError(e: any, info: any): void;
    translate(code: any, info?: any): string;
    /**
    * Returns the text of the code.
    * @desc If the language is not `en`, then it gets the text of the language.
    *
    * @returns text of that code.
    *      - if the code does not exist on text file, then it returns the code itself.
    */
    getText(code: any): string;
    getLanguage(): string;
    /**
    * Sets a language and loads the language file.
    * This will load JSON language file under `assets/lang` by default. You can change the path.
    * @desc If the input `ln` is 'en', then it will just return since `en` language is loaded by typescript by default.
    * @desc If the language is already loaded, it does not load again.
    *
    * @returns a Promise<any> on success. Otherwise error will be thrown.
    * @see README## Langulage Translation
    *
    * @code
    *          fire.setLanguage('cn')
    .catch( e => alert('Failed to load language file. ' + e.message) );
    *
    */
    setLanguage(ln: string): Promise<any>;
    /**
    *
    * Checks if the Document ID is in right format.
    *
    *
    * @param documentID Document ID
    * @returns null if there is no problem. Otherwise `ERROR CODE` will be returned.
    *
    *
    *
    */
    checkDocumentIDFormat(documentID: any): "no-document-id" | "document-id-too-long" | "document-id-cannot-cotain-slash";
    /**
     * Get a document.
     * @desc This is a general method to get a document.
     *      - It can be overriden on each module.
     *
     * @return Promise<Document Data> or Error is thrown.
     *      If there is no document by that id, then NOT_FOUND errir will be thrown.
     */
    getValidator(id: string): Promise<any>;
    get(id: string): Promise<any>;
    /**
     * Returns the document of likes/dislikes.
     *
     * @desc This gets `document of like/dislike` and used by post/comment.
     */
    private likeDocument(collectionRef);
    /**
     * This does validation for `like`, `unlike`, `dislike`, `undislike`.
     * @desc This does validator for `document of like/dislike` and used by post/comment.
     */
    private doLikeValidator(collectionRef);
    /**
     * This is a general method for `like`, `unlike`, `dislike`, `disunlike`.
     * @desc The logic is the same for `like` and `dislike` and used by post/comment.
     */
    protected doLike(collectionRef: firebase.firestore.CollectionReference): Promise<any>;
    /**
     *
     * @desc This does unlike/undislike and used by post/comment.
     */
    protected doUnlike(collectionRef: firebase.firestore.CollectionReference): Promise<any>;
    /**
     * Counts the number of Likes and saves it into `count` document.
     * @desc This cont number of `like/dislike` and used by post/comment.
     */
    private countLikes(collectionRef);
    /**
     * Returns a promise of <true|false>
     * @desc This is an Action Method.
     * @returns
     *      true if the system is already installed.
     *      false if it is not.
     */
    checkInstall(): Promise<INSTALLED>;
    /**
     * Installs the system.
     * @desc This is an Action Method.
     * @param options Options to install
     */
    install(options: any): Promise<any>;
}
