import { Base, POST, POST_CREATE, POST_EDIT, POST_DELETE, POST_PAGE_OPTIONS } from './../etc/base';
export declare class Post extends Base {
    /**
    * User class object.
    */
    private user;
    /**
    * Navigation
    * `cursor` is indicating where to load from. It is null by default.
    */
    private cursor;
    /**
    * Navigation
    * `categoryId` holds the selected category id to load and display posts from.
    * null by default
    */
    categoryId: string;
    /**
    * Posts and its IDs that have been loaded by `page()`.
    * Since object has no sequence, `pagePostIds` is holding the keys of the posts in order.
    * These are public variasble which should be used on list component to display posts.
    */
    pagePosts: {
        [id: string]: POST;
    };
    pagePostIds: Array<string>;
    /**
    * Subscribing changes for realtime update.
    */
    private _unsubscribeLikes;
    private _unsubscribePosts;
    private unsubscribePage;
    constructor();
    /**
    * Validates the input data for creating a post.
    * @desc validate the input for creating a post.
    * @desc Don't check if the category id is really exists. Normally this won't make a trouble.
    */
    private createValidator(post);
    private createSanitizer(post);
    /**
    * Creates a post.
    * @desc if `post.id` exists, then it rejects.
    * @returns Promise<POST_CREATE> with Document ID if success.
    * @since 2018-03-16 Category.numberOfPosts were removed. @see README## Client Side Coding Limitation and PUBLIC META DATA
    */
    create(post: POST): Promise<POST_CREATE>;
    /**
    * Validates edit.
    * @desc on edit, category can be empty.
    */
    private editValidator(post);
    edit(post: POST): Promise<POST_EDIT>;
    /**
    * Deletes a post.
    *
    * @desc It puts the post id under `posts_deleted` to indicate that the post has been deleted.
    * @see REAMEMD## posts_deleted collection
    * @todo test on deleting and marking.
    */
    delete(id: string): Promise<POST_DELETE>;
    /**
    * It remembers previous category for pagnation.
    * If category changes, it will clear the cursor.
    */
    private categoryChanged(category);
    /**
    * For pagination.
    *
    */
    private resetCursor(category);
    /**
    * Get posts for a page.
    * @desc if input `category` is given, then it opens a new category and gets posts for the first page.
    *    Otherwise it gets posts for next page.
    *
    * @returns true if the category has been chagned and reset. other wise false.
    */
    private resetLoadPage(category);
    /**
    * Unsubscribe all the posts.
    */
    private unsubscribePosts();
    private unsubscribeLikes();
    private subscribePostChange(post);
    /**
    * Subscribes for likes/dislikes
    * @param post post to subscribe for like, dislike
    */
    private subscribeLikes(post);
    /**
     * Gets post data to display on the page.
     *
     * 1. Build the `query` for parsing data to firebase.
     *  `collection()` - post collection super(COLLECTIONS.POST)
     *  `.where()` - category
     *  `.orderBy()` - created, desc
     *  `.startAfter()` - last doc queried
     *  `.limit()` - limits doc tobe queried.
     *
     * - Will get data based on query above.
     *  For each document `doc` parsed. We do.
     * - Gets document as object. as what `doc.data()` does.
     * - Pushes additional properties `id` amd `date` post date created.
     * - Subscribe or listen to `post changes` and `likes`.
     *
     * - Updates `cursor` based on the last `document` parsed.
     *
     * - if category changes `Post.page()` will start again and unsubscribe to posts from previous category.
     *
     * @param POST_PAGE_OPTIONS options is the options to load posts.
     * @return Promise<Array<POST>> posts that are loaded.
     *
     *
     */
    page(options: POST_PAGE_OPTIONS): Promise<Array<POST>>;
    /**
    * Listens on new post only. It does not listen for edit/delete.
    *  It subscribes added/updated/removed only after loading/displaying the post list.
    *  In this way, it prevents double display of the last post.
    */
    private subscribePostAdd(query);
    /**
    * Add a newly created post on top of post list on the page
    *  - and subscribe post changes if `settings.listenPostChange` is set to true.
    *  - and subscribe like/dislike based on the settings.
    *
    * @desc It's important to understand how `added` event fired on `onSnapshot)`.
    *
    */
    private addPostOnTop(post);
    /**
    * When listening the last post on collection in realtime, it often fires `modified` event on new docuemnt created.
    */
    private updatePost(post);
    /**
    * This method is no longer in use.
    *
    * @deprecated @see README### No post delete.
    */
    private removePost(post);
    stopLoadPage(): void;
    /**
    * Returns post docuement reference.
    */
    private post(postId);
    /**
    * Returns collection of like/dislike.
    * @param postId Post Document ID
    * @param collectionName Subcollection name
    */
    private likeColllection(postId, collectionName);
    like(postId: string): Promise<any>;
    dislike(postId: string): Promise<any>;
}
